package derez.libs;

import java.io.FileInputStream;
import java.io.IOException;
import anywheresoftware.b4a.BA.Author;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;
import anywheresoftware.b4a.keywords.DateTime;
//import anywheresoftware.b4a.objects.streams.File;
import anywheresoftware.b4a.objects.streams.File.TextReaderWrapper;

/**
 * MagneticData provides information about the magnetic parameters
 * in a location defined by its Latitude, Longitude and Altitude coordinates.
 */

@ShortName("Magnetics")
@Author("Derez")
@Version((float)1.3)
public class Magnetics
{
	private int  MAXORD ;
    private double DTR, A, B, RE, A2, B2, C2, A4, B4, C4 ;
    private double GNM, HNM, DGNM, DHNM, FLNMJ ;
    private double OTIME, OALT, OLAT ;
    private double EPOCH, TIME ;

	private double [][] C = new  double [13][13] ;
    private double [][] CD= new  double [13][13] ;
    private double [][] P = new  double [13][13] ;
    private double [][] DP = new  double [13][13] ;
    private double [][] SNORM = new  double [13][13] ;
    private double [] SP = new  double [13] ;
    private double [] CP = new  double [13] ;
    private double [] FN = new  double [13] ;
    private double [] FM  = new  double [13] ;
    private double [] PP  = new  double [13] ;
    private double [][] K  = new  double [13][13] ;
    private String cmmdir ; 
    
/**
 * Initializes the object by reading parameters from a file.
 * Do not supply File.DirAssets, If the file is provided in this directory, 
 * copy it prior to initialization to another place (for example- File.dirInternal) 
 * and then provide this directory in the initialization parameter.
 * The file is valid for the period 2015 - 2022.
 * For later use a new file should be prepared from data found in this site:
 * http://www.ngdc.noaa.gov/geomag/WMM/soft.shtml ,
 * where the algorithm is taken from.	
 */
    public void  Initialize(String DataFileDir) throws IOException 
		{
			if (DataFileDir != "" )
				cmmdir = DataFileDir ;
		}
    
    private void DoInitialize() throws IOException
	{
		int J ;
		String [] str = new String [13];
		String st;
	//	if(Assets) 
	//		st = File.Combine(File.getDirAssets(), "WMMC.COF") ;
	//	else
	//		st = File.Combine(File.getDirRootExternal() + "/Magnetics", "WMMC.COF") ;
		st = cmmdir + "/WMMC.COF" ;
		FileInputStream	inst = new FileInputStream(st) ;
		TextReaderWrapper sr = new TextReaderWrapper() ;
		sr.Initialize(inst);
		
		//       INITIALIZE CONSTANTS
  		st = sr.ReadLine().replace(" ", "") ;
  		str = st.split(",");
   		EPOCH = Double.valueOf(str[0]) ;
       	MAXORD = 12 ;
      	DTR = Math.PI/180; ;
		SP[0] = 0 ;
		CP[0] = 1 ;
		P[0][0] = 1 ;
		PP[0] = 1 ;
		DP[0][0] = 0 ;
		A = 6378.137 ;
		B = 6356.7523142 ;
		RE = 6371.2 ;
		A2 = A*A ;
		B2 = B*B ;
		C2 = A2-B2 ;
		A4 = A2*A2 ;
		B4 = B2*B2 ;
		C4 = A4-B4 ;
	//     READ WORLD MAGNETIC MODEL SPHERICAL HARMONIC COEFFICIENTS
	  	C[0][0]= 0 ;
	  	CD[0][0] = 0 ;
	  	st = sr.ReadLine().replace(" ", "") ;
   		str = st.split(",") ;
   		int check = Integer.valueOf(str[0]) ;
   		while (check < 13)
     	 {
     		int N = check ;
   			int M = Integer.valueOf(str[1]) ;
	   		GNM = Double.valueOf(str[2]) ;
	   		HNM = Double.valueOf(str[3]) ;
	   		DGNM = Double.valueOf(str[4]) ;
	   		DHNM = Double.valueOf(str[5]) ;
     		if (M <= N)
	     	{
	      		C[N][M] = GNM ;
	      		CD[N][M] = DGNM ;
		      	if(M != 0) 
		      		{
		      			C[M-1][N] = HNM ;
		      			CD[M-1][N] = DHNM ;
		     		}
	     	}
     		st = sr.ReadLine().replace(" ", "") ;
	   		str = st.split(",") ;
	   		check = Integer.valueOf(str[0]) ;
     	}
	  
   		
   		
   		
//  CONVERT SCHMIDT NORMALIZED GAUSS COEFFICIENTS TO UNNORMALIZED

		 SNORM[0][0] = 1 ;
		 for(int n=1; n<=MAXORD;n++)
		 {
		 	SNORM[n][0]= SNORM[n-1][0] * (double)(2*n-1)/(double)n ;
		   J = 2 ;
		   for(int m=0;m<=n;m++)
		   {	
		     K[n][m] = (double)((n-1)*(n-1)-m*m)/(double)((2*n-1)*(2*n-3)) ;
		     if (m > 0) 
		   	{
		   	  FLNMJ = (double)((n-m+1)* J)/(double)(n+m) ;
		   	  SNORM[n][m] = SNORM[n][m-1] * Sqrt(FLNMJ) ;
			      J=1 ;
			      C[m-1][n] = SNORM[n][m] * C[m-1][n] ;
			      CD[m-1][n] = SNORM[n][m] * CD[m-1][n] ;
		   	}
		     C[n][m] = SNORM[n][m] * C[n][m] ;
		     CD[n][m] = SNORM[n][m] * CD[n][m] ;
		   }
		   FN[n]= (double)(n+1) ;
		   FM[n] = (double)(n) ;
		}
		K[1][1] = 0  ;
		OTIME = -1000 ;
		OALT = -1000 ;
		OLAT = -1000;
		
		TIME = (double)DateTime.GetDayOfYear(DateTime.getNow()) / 365 
				+ (double)DateTime.GetYear(DateTime.getNow()) ;
	}
		
	/**
	 *  Lat and LON are in degrees, East and North positive. 
	 *  Alt in Km above sea level or the WGS84 ellipsoid.
	 *  The method returns an array of three parameters: 
	 *  -Magnetic Declination (or Variation) 
	 *  -Magnetic Inclination 
	 *  -Grid Variation - for Lat above 55 degrees or below -55 degrees.
	 *   For other cases it returns -999.
	 * 
	 */
		public String []  MagneticData(double LAT,double LON,double ALT) throws IOException
		{
		 double DT, RLON, RLAT, SRLON, SRLAT, CRLON, CRLAT ;
		 double  SRLAT2,  CRLAT2 ;
		 double Q, Q1, Q2, CT, ST, R2, R, D, CA, SA, AOR, AR ;
		 double BR, BT, BP, BPP, PAR, TEMP1, TEMP2, PARP ;
		 double BX, BY, BZ, BH ;
		 double [][] TC  = new  double [13][13] ;
		 String [] str = new String [3] ;
		 
		
		 DoInitialize() ;
		
		 
		 DT = TIME - EPOCH ;
		 
		 RLON = LON * DTR ;
		 RLAT = LAT * DTR ;
		 SRLON = Math.sin(RLON) ;
		 SRLAT = Math.sin(RLAT) ;
		 CRLON = Math.cos(RLON) ;
		 CRLAT = Math.cos(RLAT) ;
	//	 SRLON2 = SRLON*SRLON  ;
		 SRLAT2 = SRLAT*SRLAT ;
	//	 CRLON2 = CRLON*CRLON ;
		 CRLAT2 = CRLAT*CRLAT ;
		 SP[1] = SRLON ;
		 CP[1] = CRLON ;
		
		//  CONVERT FROM GEODETIC COORDS. TO SPHERICAL COORDS.
		
		 Q = Sqrt(A2 - C2 * SRLAT2) ;
		 Q1 = ALT * Q ;
		 Q2 = ((Q1+A2)/(Q1+B2))*((Q1+A2)/(Q1+B2)) ;
		 CT = SRLAT/Sqrt(Q2*CRLAT2+SRLAT2) ;
		 ST = Sqrt(1-CT*CT) ;
		 R2 = ALT*ALT+2*Q1+(A4-C4*SRLAT2)/(Q*Q) ;
		 R = Sqrt(R2) ;
		 D = Sqrt(A2*CRLAT2+B2*SRLAT2) ;
		 CA = (ALT+D)/R ;
		 SA = C2*CRLAT*SRLAT/(R*D) ;
		
		 for(int m = 2; m<=MAXORD;m++)
		  {
		  	SP[m]=SP[1]*CP[m-1]+CP[1]*SP[m-1] ;
		  	CP[m]=CP[1]*CP[m-1]-SP[1]*SP[m-1] ;
		  }
		 AOR = RE/R ;
		 AR = AOR*AOR ;
		 BR = 0 ;
		 BT = 0 ;
		 BP = 0 ;
		 BPP = 0 ;
		
		 for(int n = 1;n<=MAXORD;n++)
			{
		     AR = AR*AOR ;
		     for(int m = 0;m<=n;m++)
		 		{
		//COMPUTE UNNORMALIZED ASSOCIATED LEGENDRE POLYNOMIALS
		//AND DERIVATIVES VIA RECURSION RELATIONS
				 if ((ALT != OALT) | (LAT != OLAT)) 
				    {
				     if (n == m) 
				       {
				         P[n][m] = ST*P[n-1][m-1] ;
				         DP[n][m]=ST*DP[n-1][m-1]+CT*P[n-1][m-1] ;
				       }
				     else if (n == 1 & m == 0)
				       {			               
				         P[n][m]=CT*P[n-1][m] ;
				         DP[n][m]=CT*DP[n-1][m]-ST*P[n-1][m] ;
				       }
				     else if (n > 1 & n != m)
					   {
					     if (m > n-2)
					        P[n-2][m] = 0 ;
					     if (m > n-2)
					        DP[n-2][m] = 0 ;
					     P[n][m]=CT*P[n-1][m]-K[n][m]*P[n-2][m] ;
					     DP[n][m]=CT*DP[n-1][m]-ST*P[n-1][m]-K[n][m]*DP[n-2][m] ;
					   }
					}
		
		//  TIME ADJUST THE GAUSS COEFFICIENTS
		
		      if (TIME != OTIME)
		      {
		      	TC[n][m] = C[n][m] + DT*CD[n][m] ;
		       if (m != 0) 
		          	TC[m-1][n] = C[m-1][n] + DT * CD[m-1][n] ;
		      }
		//  ACCUMULATE TERMS OF THE SPHERICAL HARMONIC EXPANSIONS
				PAR = AR*P[n][m] ;
		       if (m == 0) 
		       {
		       	TEMP1 = TC[n][m]*CP[m] ;
		       	TEMP2 = TC[n][m]*SP[m] ;
		       }
		       else
		       {
		       	TEMP1 = TC[n][m]*CP[m]+TC[m-1][n]*SP[m] ;
		       	TEMP2 = TC[n][m]*SP[m]-TC[m-1][n]*CP[m] ;
		 		}
		       BT -=  AR * TEMP1 * DP[n][m] ;
		       BP +=  FM[m] * TEMP2 * PAR ;
		       BR +=  FN[n] * TEMP1 * PAR ;
		//SPECIAL CASE:  NORTH/SOUTH GEOGRAPHIC POLES
		       if (ST == 0 & m == 1) 
		       {
		          if (n ==1) 
		          	 PP[n]=PP[n-1] ;
		          else
		          	 PP[n]=CT*PP[n-1]-K[n][m]*PP[n-2] ;
		          PARP = AR*PP[n] ;
		          BPP +=  FM[m] * TEMP2 * PARP ;
		       }
		 }
		}
		 if (ST == 0) 
		   BP = BPP ;
		 else
		   BP = BP/ST ;
		
		//  ROTATE MAGNETIC VECTOR COMPONENTS FROM SPHERICAL TO
		//  GEODETIC COORDINATES
		
		 BX = -BT * CA - BR * SA ;
		 BY = BP ;
		 BZ = BT * SA - BR * CA ;
		//  COMPUTE DECLINATION (DEC), INCLINATION (DIP) AND
		//  TOTAL INTENSITY (TI)
		
		 BH = Sqrt(BX*BX + BY*BY) ;
	//	 double TI = Sqrt(BH*BH + BZ*BZ) ;
		 double DEC = Math.atan2(BY,BX)/DTR ;
		 double DIP = Math.atan2(BZ,BH)/DTR ;
		//  COMPUTE MAGNETIC GRID VARIATION IF THE CURRENT
		//  GEODETIC POSITION IS IN THE ARCTIC OR ANTARCTIC
		//  (I.E. LAT > +55 DEGREES OR LAT < -55 DEGREES)
		//
		//  OTHERWISE, SET MAGNETIC GRID VARIATION TO -999.0
		
		 double GV = -999.0 ;
		 if (Math.abs(LAT) > 55) 
		 {
		  	if (LAT > 0 & LON > 0)
		     	GV = DEC - LON ;
		   if (LAT > 0 & LON < 0) 
		     	GV = DEC + Math.abs(LON) ;
		   if (LAT < 0 & LON > 0) 
		     	GV = DEC + LON ;
		   if (LAT < 0 & LON < 0)
		     	GV = DEC - Math.abs(LON) ;
		   if (GV > 180) 
		     	GV = GV - 360 ;
		   if (GV < -180)
		     	GV = GV + 360 ;
		 }
		
		 OTIME = TIME ;
		 OALT = ALT ;
		 OLAT = LAT ;
		 
		 str[0] = String.valueOf(DEC) ;
		 str[1] = String.valueOf(DIP) ;
		 str[2] = String.valueOf(GV) ;
		 return str ;
		 }
		
		/**
		 * The library is taking today's date as a default. 
		 * If you need calculation for another date - use this method. 
		 */
		public void SetDate(int Day, int Month, int Year)
		{
		TIME = (double)Day / 365 + (double)Month/12 + (double)Year ;	
		}
		
		private double Sqrt(double x)
		{
			return Math.pow(x,0.5);
		}
	
}		
		//PARAMETER DESCRIPTIONS:
		//
		//  A      - SEMIMAJOR AXIS OF WGS-84 ELLIPSOID (KM)
		//  B      - SEMIMINOR AXIS OF WGS-84 ELLIPSOID (KM)
		//  RE     - MEAN RADIUS OF IAU-66 ELLIPSOID (KM)
		//  SNORM  - SCHMIDT NORMALIZATION FACTORS
		//  C      - GAUSS COEFFICIENTS OF MAIN GEOMAGNETIC MODEL (NT)
		//  CD     - GAUSS COEFFICIENTS OF SECULAR GEOMAGNETIC MODEL (NT/YR)
		// TC     - TIME ADJUSTED GEOMAGNETIC GAUSS COEFFICIENTS (NT)
		// OTIME  - TIME ON PREVIOUS CALL TO GEOMAG (YRS)
		// OALT   - GEODETIC ALTITUDE ON PREVIOUS CALL TO GEOMAG (YRS)
		// OLAT   - GEODETIC LATITUDE ON PREVIOUS CALL TO GEOMAG (DEG.)
		// TIME   - COMPUTATION TIME (YRS)                        (INPUT)
		//          (EG. 1 JULY 1995 = 1995.500)
		// ALT    - GEODETIC ALTITUDE (KM)                        (INPUT)
		// LAT   - GEODETIC LATITUDE (DEG.)                      (INPUT)
		// LON   - GEODETIC LONGITUDE (DEG.)                     (INPUT)
		// EPOCH  - BASE TIME OF GEOMAGNETIC MODEL (YRS)
		// DTR    - DEGREE TO RADIAN CONVERSION
		// SP(M)  - SINE OF (M*SPHERICAL COORD. LONGITUDE)
		// CP(M)  - COSINE OF (M*SPHERICAL COORD. LONGITUDE)
		// ST     - SINE OF (SPHERICAL COORD. LATITUDE)
		// CT     - COSINE OF (SPHERICAL COORD. LATITUDE)
		// R      - SPHERICAL COORDINATE RADIAL POSITION (KM)
		// CA     - COSINE OF SPHERICAL TO GEODETIC VECTOR ROTATION ANGLE
		// SA     - SINE OF SPHERICAL TO GEODETIC VECTOR ROTATION ANGLE
		// BR     - RADIAL COMPONENT OF GEOMAGNETIC FIELD (NT)
		// BT     - THETA COMPONENT OF GEOMAGNETIC FIELD (NT)
		// BP     - PHI COMPONENT OF GEOMAGNETIC FIELD (NT)
		// P(N,M) - ASSOCIATED LEGENDRE POLYNOMIALS (UNNORMALIZED)
		// PP(N)  - ASSOCIATED LEGENDRE POLYNOMIALS FOR M=1 (UNNORMALIZED)
		// DP(N,M)- THETA DERIVATIVE OF P(N,M) (UNNORMALIZED)
		// BX     - NORTH GEOMAGNETIC COMPONENT (NT)
		// BY     - EAST GEOMAGNETIC COMPONENT (NT)
		// BZ     - VERTICALLY DOWN GEOMAGNETIC COMPONENT (NT)
		// BH     - HORIZONTAL GEOMAGNETIC COMPONENT (NT)
		// DEC    - GEOMAGNETIC DECLINATION (DEG.)                (OUTPUT)
		//            EAST=POSITIVE ANGLES
		//            WEST=NEGATIVE ANGLES
		// DIP    - GEOMAGNETIC INCLINATION (DEG.)                (OUTPUT)
		//            DOWN=POSITIVE ANGLES
		//              UP=NEGATIVE ANGLES
		// TI     - GEOMAGNETIC TOTAL INTENSITY (NT)              (OUTPUT)
		// GV     - GEOMAGNETIC GRID VARIATION (DEG.)             (OUTPUT)
		//          REFERENCED TO GRID NORTH
		//          GRID NORTH REFERENCED TO 0 MERIDIAN
		//          OF A POLAR STEREOGRAPHIC PROJECTION
		//          (ARCTIC/ANTARCTIC ONLY)
		// MAXDEG - MAXIMUM DEGREE OF SPHERICAL HARMONIC MODEL    (INPUT)
		// MOXORD - MAXIMUM ORDER OF SPHERICAL HARMONIC MODEL
		//
		//***********************************************************************
		//
		//NOTE:  THIS VERSION OF GEOMAG USES A WMM SERIES GEOMAGNETIC
		//      FIELDS MODEL REFERENCED TO THE WGS-84 GRAVITY MODEL
//      ELLIPSOID